# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Configure pytest."""

import os
import shutil
import tempfile
from typing import List, Optional

import attr
import pytest
import requests
from assertpy import assert_that
from click.testing import CliRunner, Result
from testcontainers.core.generic import DockerContainer
from testcontainers.core.waiting_utils import wait_container_is_ready

from pynessie import cli
from pynessie.model import Content, ContentSchema, ReferenceSchema


class NessieContainer(DockerContainer):
    """Nessie test container."""

    NESSIE_PORT = 19120
    DEFAULT_IMAGE = "ghcr.io/projectnessie/nessie:latest"
    DEFAULT_TEST_IMAGE = "ghcr.io/projectnessie/nessie-unstable:latest"

    def __init__(self, image: str = DEFAULT_IMAGE) -> None:
        """Nessie test container constructor using the Nessie image tagged as 'latest'."""
        super().__init__(image=image)
        self.with_exposed_ports(NessieContainer.NESSIE_PORT)

    def get_base_url(self) -> str:
        """Get the root Nessie HTTP URL."""
        host = self.get_container_host_ip()
        port = self.get_exposed_port(NessieContainer.NESSIE_PORT)
        return f"http://{host}:{port}/"

    def get_url(self, api_version: int = 1) -> str:
        """Retrieve the Nessie API URL for the given Nessie API version, defaults to 1."""
        return f"{self.get_base_url()}api/v{api_version}"

    @wait_container_is_ready(requests.exceptions.ConnectionError, requests.exceptions.ReadTimeout)
    def _connect(self) -> None:
        response = requests.get(f"{self.get_base_url()}", timeout=1)
        response.raise_for_status()

    def start(self) -> "NessieContainer":
        """Starts the Nessie container, waits until Nessie is ready."""
        super().start()
        self._connect()
        return self


@attr.dataclass
class NessieTestConfig:
    """Test configs for pynessie tests."""

    config_dir: str
    cleanup: bool

    nessie_container: Optional[NessieContainer]


nessie_test_config: NessieTestConfig = NessieTestConfig("", False, None)


def pytest_configure(config):  # noqa
    """Configure pytest."""
    config.addinivalue_line("markers", "doc: mark as end-to-end test.")


def pytest_sessionstart(session: pytest.Session) -> None:
    """Setup a fresh temporary config directory for tests."""
    nessie_test_config.config_dir = tempfile.mkdtemp() + "/"
    # Instruct Confuse to keep Nessie config file in the temp location:
    os.environ["NESSIEDIR"] = nessie_test_config.config_dir


def pytest_sessionfinish(session: pytest.Session, exitstatus: int) -> None:
    """Remove temporary config directory."""
    shutil.rmtree(nessie_test_config.config_dir)

    if nessie_test_config.nessie_container is not None:
        nessie_test_config.nessie_container.stop()


def pytest_runtest_setup(item: pytest.Item) -> None:
    """Starts a Nessie container, if needed."""
    if nessie_test_config.nessie_container is None and item.get_closest_marker("nessieserver") is not None:
        image = os.environ.get("NESSIE_TEST_IMAGE", NessieContainer.DEFAULT_TEST_IMAGE)
        nessie_test_config.nessie_container = NessieContainer(image=image)
        nessie_test_config.nessie_container.start()

        os.environ["NESSIE_ENDPOINT"] = nessie_test_config.nessie_container.get_url(1)


def pytest_runtest_teardown(item: pytest.Item, nextitem: Optional[pytest.Item]) -> None:
    """Stops the Nessie container, if one was started."""
    if item.get_closest_marker("nessieserver") is not None and nessie_test_config.nessie_container is not None:
        reset_nessie_server_state()


def execute_cli_command_raw(args: List[str], input_data: Optional[str] = None, ret_val: int = 0) -> Result:
    """Execute a Nessie CLI command."""
    result = CliRunner().invoke(cli.cli, args, input=input_data)
    if result.exit_code != ret_val:
        print(f"Nessie CLI exited with unexpected code {result.exit_code}, expected {ret_val}")
        print(result.stdout)
        print(result.stderr)
        print(result)  # exception
    assert_that(result.exit_code).is_equal_to(ret_val)
    return result


def execute_cli_command(args: List[str], input_data: Optional[str] = None, ret_val: int = 0) -> str:
    """Execute a Nessie CLI command and return its STDOUT."""
    return execute_cli_command_raw(args, input_data=input_data, ret_val=ret_val).stdout


def ref_hash(ref: str) -> str:
    """Get the hash for a reference."""
    refs = ReferenceSchema().loads(execute_cli_command(["--json", "branch", "-l"]), many=True)
    return next(i.hash_ for i in refs if i.name == ref)


def make_commit(
    key: str, table: Content, branch: str, head_hash: Optional[str] = None, message: str = "test message", author: str = "nessie test"
) -> None:
    """Make commit through Nessie CLI."""
    if not head_hash:
        refs = {i.name: i.hash_ for i in ReferenceSchema().loads(execute_cli_command(["--json", "branch"]), many=True)}
        head_hash = refs[branch]
    assert isinstance(head_hash, str)
    execute_cli_command(
        ["content", "commit", "--stdin", key, "--ref", branch, "-m", message, "-c", head_hash, "--author", author],
        input_data=ContentSchema().dumps(table),
    )


def reset_nessie_server_state() -> None:
    """Resets the Nessie Server to an initial, clean state for testing."""
    # Delete all branches except main
    branches = ReferenceSchema().loads(execute_cli_command(["--json", "branch"]), many=True)
    for branch in branches:
        if branch.name != "main":
            execute_cli_command(["branch", "-d", branch.name])

    # Delete all tags
    tags = ReferenceSchema().loads(execute_cli_command(["--json", "tag"]), many=True)
    for tag in tags:
        execute_cli_command(["tag", "-d", tag.name])

    # Note: This hash should match the java constant AbstractDatabaseAdapter.NO_ANCESTOR
    no_ancestor_hash = "2e1cfa82b035c26cbbbdae632cea070514eb8b773f616aaeaf668e2f0be8f10d"

    # Reset the main branch to the "root" (a.k.a. no ancestor) hash
    execute_cli_command(["branch", "--force", "-o", no_ancestor_hash, "main", "main"])

    # Verify that the main branch has been reset
    branches = ReferenceSchema().loads(execute_cli_command(["--json", "branch"]), many=True)
    assert_that(branches).is_length(1)
    assert_that(branches[0].name).is_equal_to("main")
    assert_that(branches[0].hash_).is_equal_to(no_ancestor_hash)



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""merge CLI command."""

import click
from click import UsageError

from pynessie.cli_common_context import ContextObject, MutuallyExclusiveOption
from pynessie.decorators import error_handler, pass_client, validate_reference
from pynessie.model import MergeResponseSchema


@click.command("merge")
@click.option("-b", "--branch", "ref", help="branch to merge onto. If not supplied the default branch from config is used")
@click.argument("from_ref", nargs=1, required=False)
@click.option(
    "-f",
    "--force",
    cls=MutuallyExclusiveOption,
    is_flag=True,
    mutually_exclusive=["condition"],
    help="force branch assignment",
)
@click.option(
    "-c",
    "--condition",
    "expected_hash",
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["force"],
    help="Expected hash. Only perform the action if the branch currently points to the hash specified by this option.",
)
@click.option("-o", "--hash-on-ref", help="Hash on merge-from-reference")
@pass_client
@error_handler
@validate_reference
def merge(ctx: ContextObject, ref: str, force: bool, expected_hash: str, hash_on_ref: str, from_ref: str) -> None:
    """Merge FROM_REF into another branch.

    FROM_REF can be a hash or branch.

    Examples:

        nessie merge -c 12345678abcdef dev -> merge dev to default branch with default branch's
    expected hash '12345678abcdef'

        nessie merge -b main -c 12345678abcdef dev -> merge dev to a branch named main with main branch's
    expected hash '12345678abcdef'

        nessie merge -b main -o 56781234abcdef -c 12345678abcdef dev -> merge dev at hash-on-ref '56781234abcdef' to
    main branch with main branch's expected hash '12345678abcdef'

        nessie merge -f -b main dev -> forcefully merge dev to a branch named main
    """
    if not force and not expected_hash:
        raise UsageError(
            """Either condition or force must be set. Condition should be set to a valid hash for concurrency
            control or force to ignore current state of Nessie Store."""
        )
    merge_response = ctx.nessie.merge(from_ref, ref, hash_on_ref, expected_hash)
    if ctx.json:
        click.echo(MergeResponseSchema().dumps(merge_response))
    elif merge_response.was_applied:
        click.echo(f"Merged {from_ref} onto {merge_response.target_branch} (was on {merge_response.effective_target_hash} before merge)")
        if merge_response.common_ancestor:
            click.echo(f"Identified merge base commit {merge_response.common_ancestor}")
        if merge_response.resultant_target_hash:
            click.echo(f"Resultant hash on {merge_response.target_branch} after merge: {merge_response.resultant_target_hash}")
    else:
        click.echo(f"Nothing merged from {from_ref} onto {merge_response.target_branch} (still on {merge_response.effective_target_hash})")
        if merge_response.common_ancestor:
            click.echo(f"Identified merge base commit {merge_response.common_ancestor}")
        if merge_response.resultant_target_hash:
            click.echo(f"Current, unchanged hash on {merge_response.target_branch} after merge: {merge_response.resultant_target_hash}")



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Top-level package for Nessie CLI commands."""

from pynessie.commands.branch import branch_
from pynessie.commands.cherry_pick import cherry_pick
from pynessie.commands.config import config
from pynessie.commands.content import content
from pynessie.commands.diff import diff
from pynessie.commands.log import log
from pynessie.commands.merge import merge
from pynessie.commands.remote import remote
from pynessie.commands.tag import tag

__all__ = ["remote", "tag", "branch_", "cherry_pick", "config", "log", "merge", "content", "diff"]



# Copyright (C) 2020 Dremio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Tests for `pynessie` package."""
import pytest

from pynessie import init
from pynessie.client._endpoints import _sanitize_url
from pynessie.error import NessieConflictException
from pynessie.model import Branch, Entries


@pytest.mark.nessieserver
def test_client_interface_e2e() -> None:
    """Test client object against live server."""
    client = init()
    assert isinstance(client.get_base_url(), str)
    assert client.get_base_url().startswith("http://localhost:")
    assert client.get_base_url().endswith("/api/v1")
    references = client.list_references().references
    assert len(references) == 1
    assert references[0] == Branch("main", references[0].hash_)
    main_name = references[0].name
    main_commit = references[0].hash_
    with pytest.raises(NessieConflictException):
        client.create_branch("main", "main", client.get_reference(None).hash_)
    created_reference = client.create_branch("test", main_name, main_commit)
    created_reference_with_slash = client.create_branch("test/branch/name", main_name, main_commit)
    references = client.list_references().references
    assert len(references) == 3
    assert next(i for i in references if i.name == "main") == Branch("main", main_commit)
    assert next(i for i in references if i.name == "test") == Branch("test", main_commit)
    assert next(i for i in references if i.name == "test/branch/name") == Branch("test/branch/name", main_commit)
    # Testing multiple references
    ref_to_test = {
        "test": created_reference,
        "test/branch/name": created_reference_with_slash,
    }
    for ref_name, create_branch_ref in ref_to_test.items():
        reference = client.get_reference(ref_name)
        assert create_branch_ref == reference
        assert isinstance(reference.hash_, str)
        tables = client.list_keys(reference.name, reference.hash_)
        assert isinstance(tables, Entries)
        assert len(tables.entries) == 0
    assert isinstance(main_commit, str)
    client.delete_branch("test", main_commit)
    references = client.list_references().references
    assert len(references) == 2
    client.delete_branch("test/branch/name", main_commit)
    references = client.list_references().references
    assert len(references) == 1


def test_client_sanitize_url() -> None:
    """Test sanitization of URLs."""
    client = init()
    base_url = client.get_base_url()
    assert _sanitize_url(base_url) == base_url
    assert _sanitize_url(base_url + "/trees/tree") == base_url + "/trees/tree"
    assert _sanitize_url(base_url + "/trees/tree/{}", "my tag with spaces") == base_url + "/trees/tree/my%20tag%20with%20spaces"
    assert _sanitize_url(base_url + "/trees/tree/{}", "my/tag with mixed.types") == base_url + "/trees/tree/my%2Ftag%20with%20mixed.types"
    assert (
        _sanitize_url(base_url + "/trees/tree/{}/{}", "tag/name", "other/string@with-at")
        == base_url + "/trees/tree/tag%2Fname/other%2Fstring%40with-at"
    )
